//! Classes for scheduling
//!
//! # Examples
//!
//! ```
//! use pretty::scheduler::{Scheduler, BinaryHeapScheduler, ScheduledTaskId, ScheduledTask};
//! use pretty::resultconsumer::Runnable;
//! use std::time::{Instant, Duration};
//!
//! // create a scheduler of type BinaryHeapScheduler
//! let mut scheduler : Box<Scheduler> = Box::new(BinaryHeapScheduler::new());
//!
//! let now = Instant::now();
//!
//! // lets create 3 tasks
//! let task1_instant = now + Duration::from_millis(1000);
//! let task1_id = scheduler.add(ScheduledTask::from_instant(task1_instant, Runnable::new(move || { println!("scheduled task 1");})));
//!
//! let task2_instant = now + Duration::from_millis(2000);
//! let task2_id = scheduler.add(ScheduledTask::from_instant(task2_instant, Runnable::new(move || { println!("scheduled task 2");})));
//!
//! let task3_instant = now + Duration::from_millis(500);
//! let task3_id = scheduler.add(ScheduledTask::from_instant(task3_instant, Runnable::new(move || { println!("scheduled task 3");})));
//!
//! // they all have differnt ids
//! assert_ne!(task1_id, task2_id);
//! assert_ne!(task1_id, task3_id);
//! assert_ne!(task2_id, task3_id);
//!
//! // scheduler.next_instant() gives you the instant the next task is supposed to run (task3 in this case)
//! assert_eq!(Some(task3_instant), scheduler.next_instant());
//!
//! // pop_expired removes an expired task from the scheduler
//! assert_eq!(true, scheduler.pop_expired(now).is_none());
//! assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(499)).is_none());
//!
//! // but only if its time has come
//! assert_eq!(task3_id, scheduler.pop_expired(now + Duration::from_millis(500)).unwrap().id);
//!
//! // next task should be task1
//! assert_eq!(Some(task1_instant), scheduler.next_instant());
//!
//! // we remove it
//! scheduler.remove(task1_id);
//!
//! // next task should be task2
//! assert_eq!(Some(task2_instant), scheduler.next_instant());
//! assert_eq!(true, scheduler.pop_expired(now).is_none());
//! assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(1999)).is_none());
//! assert_eq!(task2_id, scheduler.pop_expired(now + Duration::from_millis(2000)).unwrap().id);
//!
//! // no next task
//! assert_eq!(None, scheduler.next_instant());
//! ```
extern crate copyless;

use std::cmp::{self, Ord};
use std::collections::{BinaryHeap, HashMap};
use std::fmt;
use std::sync::atomic::{AtomicUsize, Ordering};
use std::time::{Duration, Instant};

use copyless::BoxHelper;

/// a runnable once task, see https://doc.rust-lang.org/book/second-edition/ch20-05-sending-requests-via-channels.html why this type is necessary and we cannot simply use FnBox
#[allow(dead_code)]
pub struct Runnable {
    runnable_box: Box<RunnableBox + Send>
}

impl Runnable {
    #[allow(dead_code)]
    pub fn new_box<F>(func: F) -> Runnable
        where
            F: FnOnce() + Send + 'static {
        Runnable { runnable_box: Box::new(func) }
    }
    #[allow(dead_code)]
    pub fn new_copyless<F>(func: F) -> Runnable
        where
            F: FnOnce() + Send + 'static {
        Runnable { runnable_box: Box::alloc().init(func) }
    }
    #[allow(dead_code)]
    pub fn run(self) {
        self.runnable_box.run();
    }
}

impl fmt::Debug for Runnable {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Runnable")
    }
}

trait RunnableBox {
    fn run(self: Box<Self>);
}

impl<F: FnOnce()> RunnableBox for F {
    fn run(self: Box<F>) {
        (*self)()
    }
}


/// a basic type for task ids
#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct ScheduledTaskId(usize);

static SCHEDULED_TASK_ID_COUNTER: AtomicUsize = AtomicUsize::new(0);

impl ScheduledTaskId {
    pub fn new() -> ScheduledTaskId {
        ScheduledTaskId(SCHEDULED_TASK_ID_COUNTER.fetch_add(1, Ordering::Relaxed) + 1)
    }
}

impl Default for ScheduledTaskId {
    fn default() -> Self {
        ScheduledTaskId::new()
    }
}


impl From<usize> for ScheduledTaskId {
    fn from(val: usize) -> ScheduledTaskId {
        ScheduledTaskId(val)
    }
}

impl From<ScheduledTaskId> for usize {
    fn from(val: ScheduledTaskId) -> usize {
        val.0
    }
}

/// A container of a scheduled task, consisting of an id, an instant and a task to run
#[derive(Debug)]
pub struct ScheduledTask {
    pub id: ScheduledTaskId,
    pub instant: Instant,
    pub task: Runnable,
}

impl ScheduledTask {
    pub fn from_instant(instant: Instant, task: Runnable) -> ScheduledTask {
        ScheduledTask::from_id(ScheduledTaskId::new(), instant, task)
    }

    pub fn from_delay(delay: Duration, task: Runnable) -> ScheduledTask {
        ScheduledTask::from_id(ScheduledTaskId::new(), Instant::now() + delay, task)
    }

    pub fn from_id(id: ScheduledTaskId, instant: Instant, task: Runnable) -> ScheduledTask {
        ScheduledTask {
            id,
            instant,
            task,
        }
    }
}

#[allow(dead_code)]
#[derive(Debug)]
pub enum SchedulerOperation {
    Add(ScheduledTask),
    Remove(ScheduledTaskId),
}


pub trait Scheduler {
    /// return the duration until the next task should be executed
    fn duration_to_next(&self) -> Option<Duration>;

    /// return the instant the next task should be executed
    fn next_instant(&self) -> Option<Instant>;

    /// return the task to be executed next, but only if it is expired (i.e. its instant is < now)
    fn pop_expired(&mut self, now: Instant) -> Option<ScheduledTask>;

    /// remove/unschedule a task
    fn remove(&mut self, id: ScheduledTaskId) -> Option<ScheduledTask>;

    /// add a new task
    fn add(&mut self, task: ScheduledTask) -> ScheduledTaskId;
    /// clear all
    fn clear(&mut self);
}

#[derive(Copy, Clone, Eq, PartialEq)]
struct TaskInstant {
    instant: Instant,
    id: ScheduledTaskId,
}

impl Ord for TaskInstant {
    fn cmp(&self, other: &TaskInstant) -> cmp::Ordering {
        // Notice that the we flip the ordering on costs.
        // In case of a tie we compare positions - this step is necessary
        // to make implementations of `PartialEq` and `Ord` consistent.
        other.instant.cmp(&self.instant)
            .then_with(|| other.id.cmp(&self.id))
    }
}

// `PartialOrd` needs to be implemented as well.
impl PartialOrd for TaskInstant {
    fn partial_cmp(&self, other: &TaskInstant) -> Option<cmp::Ordering> {
        Some(self.cmp(other))
    }
}


#[allow(dead_code)]
pub struct BinaryHeapScheduler {
    #[allow(dead_code)]
    timers: HashMap<usize, ScheduledTask>,
    instants: BinaryHeap<TaskInstant>,
    #[allow(dead_code)]
    next_instant: Option<Instant>,

}

impl BinaryHeapScheduler {
    #[allow(dead_code)]
    pub fn new() -> BinaryHeapScheduler {
        BinaryHeapScheduler {
            timers: HashMap::new(),
            instants: BinaryHeap::new(),
            next_instant: None,
        }
    }
    fn find_next_instant(&mut self) -> Option<Instant> {
        loop {
            match self.instants.peek() {
                None => return None,
                Some(task_instant) => if self.timers.contains_key(&task_instant.id.0) {
                    return Some(task_instant.instant);
                }
            }
            self.instants.pop();
        };
    }
    #[allow(dead_code)]
    fn update_next(&mut self) {
        self.next_instant = self.find_next_instant();
    }
}

impl Default for BinaryHeapScheduler {
    fn default() -> Self {
        BinaryHeapScheduler::new()
    }
}

impl Scheduler for BinaryHeapScheduler {
    #[allow(dead_code)]
    fn duration_to_next(&self) -> Option<Duration> {
        self.next_instant.map(|instant| {
            let now = Instant::now();
            if instant < now {
                Duration::new(0, 0)
            } else {
                instant - now
            }
        })
    }
    fn next_instant(&self) -> Option<Instant> {
        self.next_instant
    }

    #[allow(dead_code)]
    fn pop_expired(&mut self, now: Instant) -> Option<ScheduledTask> {
        loop {
            match self.next_instant {
                None => { return None; }
                Some(instant) => if instant > now { return None; },
            }

            let task_id =
                match self.instants.pop() {
                    Some(task_instant) => task_instant.id,
                    _ => return None
                };

            self.update_next();

            match self.remove(task_id) {

                // retry if already removed
                None => {
                    continue;
                }
                // otherwise use
                Some(task) => {
                    return Some(task);
                }
            }
        }
    }
    #[allow(dead_code)]
    fn remove(&mut self, id: ScheduledTaskId) -> Option<ScheduledTask> {
        // we do not remove since instants a binary heap cannot remove specific elements
        let option = self.timers.remove(&id.0);
        if option.is_some() {
            self.update_next();
        }
        option
    }
    #[allow(dead_code)]
    fn add(&mut self, task: ScheduledTask) -> ScheduledTaskId {
        let id = task.id;
        self.instants.push(TaskInstant { instant: task.instant, id });
        self.timers.insert(task.id.0, task);
        self.update_next();
        id
    }

    fn clear(&mut self) {
        self.instants.clear();
        self.timers.clear();
        self.next_instant = None;
    }
}


#[cfg(test)]
mod tests {
    use super::BinaryHeapScheduler;
    use super::Scheduler;

    #[test]
    fn test_copyless_doesntwork() {
        let mut scheduler = BinaryHeapScheduler::new();
        use super::Runnable;
        use super::ScheduledTask;
        use std::time::{Duration, Instant};

        let now = Instant::now();

        // lets create 3 tasks
        let task1_instant = now + Duration::from_millis(1000);
        let task1_id = scheduler.add(ScheduledTask::from_instant(task1_instant, Runnable::new_copyless(move || { println!("scheduled task 1"); })));

        let task2_instant = now + Duration::from_millis(2000);
        let task2_id = scheduler.add(ScheduledTask::from_instant(task2_instant, Runnable::new_copyless(move || { println!("scheduled task 2"); })));

        let task3_instant = now + Duration::from_millis(500);
        let task3_id = scheduler.add(ScheduledTask::from_instant(task3_instant, Runnable::new_copyless(move || { println!("scheduled task 3"); })));

        // they all have differnt ids
        assert_ne!(task1_id, task2_id);
        assert_ne!(task1_id, task3_id);
        assert_ne!(task2_id, task3_id);

        // scheduler.next_instant() gives you the instant the next task is supposed to run (task3 in this case)
        assert_eq!(Some(task3_instant), scheduler.next_instant());

        // pop_expired removes an expired task from the scheduler
        assert_eq!(true, scheduler.pop_expired(now).is_none());
        assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(499)).is_none());

        // but only if its time has come
        assert_eq!(task3_id, scheduler.pop_expired(task3_instant).unwrap().id);

        // next task should be task1
        assert_eq!(Some(task1_instant), scheduler.next_instant());

        // we remove it
        scheduler.remove(task1_id);

        // next task should be task2
        assert_eq!(Some(task2_instant), scheduler.next_instant());
        assert_eq!(true, scheduler.pop_expired(now).is_none());
        assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(1999)).is_none());
        assert_eq!(task2_id, scheduler.pop_expired(now + Duration::from_millis(2000)).unwrap().id);

        // no next task
        assert_eq!(None, scheduler.next_instant());
    }


    #[test]
    fn test_copyless_works() {
        let mut scheduler = BinaryHeapScheduler::new();

        use super::Runnable;
        use super::ScheduledTask;
        use std::time::{Duration, Instant};

        let now = Instant::now();

        // lets create 3 tasks
        let task1_instant = now + Duration::from_millis(1000);
        let task1_id = scheduler.add(ScheduledTask::from_instant(task1_instant, Runnable::new_copyless(move || { println!("scheduled task 1 {:?}", now); })));

        let task2_instant = now + Duration::from_millis(2000);
        let task2_id = scheduler.add(ScheduledTask::from_instant(task2_instant, Runnable::new_copyless(move || { println!("scheduled task 2 {:?}", now); })));

        let task3_instant = now + Duration::from_millis(500);
        let task3_id = scheduler.add(ScheduledTask::from_instant(task3_instant, Runnable::new_copyless(move || { println!("scheduled task 3 {:?}", now); })));

        // they all have differnt ids
        assert_ne!(task1_id, task2_id);
        assert_ne!(task1_id, task3_id);
        assert_ne!(task2_id, task3_id);

        // scheduler.next_instant() gives you the instant the next task is supposed to run (task3 in this case)
        assert_eq!(Some(task3_instant), scheduler.next_instant());

        // pop_expired removes an expired task from the scheduler
        assert_eq!(true, scheduler.pop_expired(now).is_none());
        assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(499)).is_none());

        // but only if its time has come
        assert_eq!(task3_id, scheduler.pop_expired(task3_instant).unwrap().id);

        // next task should be task1
        assert_eq!(Some(task1_instant), scheduler.next_instant());

        // we remove it
        scheduler.remove(task1_id);

        // next task should be task2
        assert_eq!(Some(task2_instant), scheduler.next_instant());
        assert_eq!(true, scheduler.pop_expired(now).is_none());
        assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(1999)).is_none());
        assert_eq!(task2_id, scheduler.pop_expired(now + Duration::from_millis(2000)).unwrap().id);

        // no next task
        assert_eq!(None, scheduler.next_instant());
    }

    #[test]
    fn test_box_new() {
        let mut scheduler = BinaryHeapScheduler::new();

        use super::Runnable;
        use super::ScheduledTask;
        use std::time::{Duration, Instant};

        let now = Instant::now();

        // lets create 3 tasks
        let task1_instant = now + Duration::from_millis(1000);
        let task1_id = scheduler.add(ScheduledTask::from_instant(task1_instant, Runnable::new_box(move || { println!("scheduled task 1"); })));

        let task2_instant = now + Duration::from_millis(2000);
        let task2_id = scheduler.add(ScheduledTask::from_instant(task2_instant, Runnable::new_box(move || { println!("scheduled task 2"); })));

        let task3_instant = now + Duration::from_millis(500);
        let task3_id = scheduler.add(ScheduledTask::from_instant(task3_instant, Runnable::new_box(move || { println!("scheduled task 3"); })));

        // they all have differnt ids
        assert_ne!(task1_id, task2_id);
        assert_ne!(task1_id, task3_id);
        assert_ne!(task2_id, task3_id);

        // scheduler.next_instant() gives you the instant the next task is supposed to run (task3 in this case)
        assert_eq!(Some(task3_instant), scheduler.next_instant());

        // pop_expired removes an expired task from the scheduler
        assert_eq!(true, scheduler.pop_expired(now).is_none());
        assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(499)).is_none());

        // but only if its time has come
        assert_eq!(task3_id, scheduler.pop_expired(task3_instant).unwrap().id);

        // next task should be task1
        assert_eq!(Some(task1_instant), scheduler.next_instant());

        // we remove it
        scheduler.remove(task1_id);

        // next task should be task2
        assert_eq!(Some(task2_instant), scheduler.next_instant());
        assert_eq!(true, scheduler.pop_expired(now).is_none());
        assert_eq!(true, scheduler.pop_expired(now + Duration::from_millis(1999)).is_none());
        assert_eq!(task2_id, scheduler.pop_expired(now + Duration::from_millis(2000)).unwrap().id);

        // no next task
        assert_eq!(None, scheduler.next_instant());
    }
}
